@@ -123,6 +123,17 @@ check_env() {
 	modprobe multipath 2> /dev/null
 	grep -sq 'Personalities : .*multipath' /proc/mdstat &&
 		MULTIPATH="yes"
+	if [ "$MULTIPATH" != "yes" ]; then
+		echo "test: skipping tests for multipath, which is removed in upstream 6.8+ kernels"
+	fi
+
+	# Check whether to run linear tests
+	modprobe linear 2> /dev/null
+	grep -sq 'Personalities : .*linear' /proc/mdstat &&
+		LINEAR="yes"
+	if [ "$LINEAR" != "yes" ]; then
+		echo "test: skipping tests for linear, which is removed in upstream 6.8+ kernels"
+	fi
 }
 
 do_setup() {
@@ -202,7 +213,6 @@ do_setup() {
 	path1=$dev7
 	ulimit -c unlimited
 	[ -f /proc/mdstat ] || modprobe md_mod
-	echo 2000 > /proc/sys/dev/raid/speed_limit_max
 	echo 0 > /sys/module/md_mod/parameters/start_ro
 }
 
